%{
#include <stdio.h>
#include "scheme_interp.h"
extern int yylex();

exp *parse_res;

void yyerror (char *s)  /* Called by yyparse on error */
{
  printf ("%s\n", s);
  exit(3);
}

cons_cell *cons(void *car, cons_cell *cdr)
{
  cons_cell *c = (cons_cell*)malloc(sizeof(cons_cell));
  c->car = car;
  c->cdr = cdr;
  return c;
}

int len(cons_cell *c)
{
  int i = 0;
  while (c != NULL)
  {
    i++;
    c = c->cdr;
  }
  return i;
}

void **tovec(cons_cell *c)
{
  cons_cell *tmp;
  int i = len(c);
  void **res = (void**)ALLOC(sizeof(void*) * i);
  i = 0;
  while (c != NULL)
  {
    res[i] = c->car;
    tmp = c;
    c = c->cdr;
    free(tmp);
    i++;
  }
  return res;
}

void **tovec1(cons_cell *c)
{
  int i = len(c);
  void **res = (void**)ALLOC(sizeof(void*) * i);
  i = 0;
  while (c != NULL)
  {
    res[i] = ((cons_cell*)c->car)->car;
    c = c->cdr;
    i++;
  }
  return res;
}

void **tovec2(cons_cell *c)
{
  cons_cell *tmp;
  int i = len(c);
  void **res = (void**)ALLOC(sizeof(void*) * i);
  i = 0;
  while (c != NULL)
  {
    res[i] = ((cons_cell*)c->car)->cdr;
    tmp = c;
    c = c->cdr;
    free(tmp->car);
    free(tmp);
    i++;
  }
  return res;
}


%}

%union
{
  int num;
  char *str;
  exp *e;
  cons_cell *p;
}

%token <num> NUM
%token <str> ID
%token NLL HASHF VOID LAMBDA LET LETREC IF BGIN SET AND OR NOT

%type <e> ex p
%type <p> ids ex_list let_binds

%%

p: ex { parse_res = $1; }
;

ex: NUM { $$ = make_num_exp($1); }
  | NLL { $$ = make_null_exp(); }
  | HASHF { $$ = make_false_exp(); }
  | VOID { $$ = make_void_exp(); }
  | '(' LAMBDA '(' ids ')' ex ')'
       { $$ = make_lambda_exp(len($4), (char**)tovec($4), $6); }
  | ID { $$ = make_var_exp($1); }
  | '(' LET '(' let_binds ')' ex ')' 
       { $$ = make_let_exp(len($4), (char**)tovec1($4), (exp**)tovec2($4), $6); }
  | '(' LETREC '(' let_binds ')' ex ')'
       { $$ = make_letrec_exp(len($4), (char**)tovec1($4), (exp**)tovec2($4), $6); }
  | '(' ex ex_list ')' { $$ = make_app_exp($2, len($3), (exp**)tovec($3)); }
  | '(' IF ex ex ex ')' { $$ = make_if_exp($3, $4, $5); }
  | '(' BGIN ex_list ')' { $$ = make_begin_exp(len($3), (exp**)tovec($3)); }
  | '(' SET ID ex ')' { $$ = make_setbang_exp($3, $4); }
  | '(' AND ex_list ')' { $$ = make_and_exp(len($3), (exp**)tovec($3)); }
  | '(' OR ex_list ')' { $$ = make_or_exp(len($3), (exp**)tovec($3)); }
  | '(' NOT ex ')' { $$ = make_not_exp($3); }
;
ids:    { $$ = NULL; }
  | ID ids  { $$ = cons($1, $2); }
;
let_binds:  { $$ = NULL; }
  | '(' ID ex ')' let_binds  { $$ = cons(cons($2, (cons_cell*)$3) , $5); }
;
ex_list:   { $$ = NULL; }
  | ex ex_list  { $$ = cons($1, $2); }
;
  
%%
