#include "scheme_interp.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

extern FILE *yyin;
extern void yyparse();
extern exp *parse_res;

exp *parse_file(char *f)
{
  yyin = fopen(f, "r");
  yyparse();
  return parse_res;
}

static cexp *compute_lex_addr(char *var, env *e)
{
  int dist = 0;
  while (1)
  {
    if (e->tag == EMPTY_ENV)
    {
      printf("Free variable %s.\n", var);
      exit(1);
    }
    else
    {
      int i;
      char **syms = ((extended_env*)e)->syms;
      int len = ((extended_env*)e)->num_bindings;
      for (i = 0; i < len; ++i)
	if (strcmp(var, syms[i]) == 0)
	  return make_lexvar_cexp(dist, i);
      e = ((extended_env*)e)->next;
      dist++;
    }
  }
}

static inline exp *compile_let(int num_bounds, char **ids, exp **bound_exps,
			       exp *body)
{
  exp *f = make_lambda_exp(num_bounds, ids, body);
  return make_app_exp(f, num_bounds, bound_exps);
}

static inline exp *compile_letrec(int num_bounds, char **ids, exp **bound_exps,
				  exp *body)
{
  int i;
  exp **sets, *be;
  exp *dummy = make_void_exp();
  exp **dummies = (exp**)ALLOC(sizeof(exp*) * num_bounds);
  for (i = 0; i < num_bounds; ++i)
    dummies[i] = dummy;
  sets = (exp**)ALLOC(sizeof(setbang_exp*) * (num_bounds + 1));
  for (i = 0; i < num_bounds; ++i)
    sets[i] = make_setbang_exp(ids[i], bound_exps[i]);
  sets[i] = body;
  be = make_begin_exp(num_bounds + 1, sets);
  return make_let_exp(num_bounds, ids, dummies, be);
}

static inline exp *compile_begin(int num_exps, exp **exps)
{
  int i;
  exp *f;
  char **parms = (char**)ALLOC(sizeof(char*) * (num_exps - 1));
  for (i = 0; i < num_exps - 1; ++i)
    parms[i] = "";
  f = make_lambda_exp(num_exps - 1, parms, exps[num_exps - 1]);

  return make_app_exp(f, num_exps - 1, exps);
}

cexp *compile(exp *prog, env *e)
{
  switch (prog->tag)
  {
  case NUM_EXP:
    return make_val_cexp(make_number_val(((num_exp*)prog)->n));
  case NULL_EXP:
    return make_val_cexp(make_null_val());
  case FALSE_EXP:
    return make_val_cexp(make_false_val());
  case VOID_EXP:
    return make_val_cexp(make_void_val());
  case LAMBDA_EXP:
  {
    env *new_env = make_extended_env(((lambda_exp*)prog)->num_parms,
				     ((lambda_exp*)prog)->parms,
				     NULL,
				     e);
    return make_lambda_cexp(((lambda_exp*)prog)->num_parms,
			    compile(((lambda_exp*)prog)->body, new_env));
  }
  case VAR_EXP:
    return compute_lex_addr(((var_exp*)prog)->var, e);
  case LET_EXP:
    return compile(compile_let(((let_exp*)prog)->num_bindings,
			       ((let_exp*)prog)->ids,
			       ((let_exp*)prog)->bound_exps,
			       ((let_exp*)prog)->body),
		   e);
  case LETREC_EXP:
    return compile(compile_letrec(((letrec_exp*)prog)->num_bindings,
				  ((letrec_exp*)prog)->ids,
				  ((letrec_exp*)prog)->bound_exps,
				  ((letrec_exp*)prog)->body),
		   e);
  case APP_EXP:
  {
    int i;
    int num_args = ((app_exp*)prog)->num_args;
    cexp **args = (cexp**)ALLOC(sizeof(cexp*) * num_args);
    for (i = 0; i < num_args; ++i)
      args[i] = compile(((app_exp*)prog)->args[i], e);
    return make_app_cexp(compile(((app_exp*)prog)->fun, e),
			 num_args, args);			 
  }
  case IF_EXP:
    return make_if_cexp(compile(((if_exp*)prog)->test, e),
			compile(((if_exp*)prog)->t_br, e),
			compile(((if_exp*)prog)->f_br, e));
  case BEGIN_EXP:
    return compile(compile_begin(((begin_exp*)prog)->num_exps,
				 ((begin_exp*)prog)->exps),
		   e);
  case SETBANG_EXP:
  {
    cexp *d = compute_lex_addr(((setbang_exp*)prog)->var, e);
    return make_lexsetbang_cexp(((lexvar_cexp*)d)->dist,
				((lexvar_cexp*)d)->offset,
				compile(((setbang_exp*)prog)->val, e));
  }
  case AND_EXP:
  {
    int i;
    int num_args = ((and_exp*)prog)->num_args;
    if (num_args == 0)
      return make_val_cexp(make_number_val(0));
    else
    {
      cexp **args = (cexp**)ALLOC(sizeof(cexp*) * num_args);
      for (i = 0; i < num_args; ++i)
	args[i] = compile(((and_exp*)prog)->args[i], e);
      return make_and_cexp(num_args, args);
    }
  }
  case OR_EXP:
  {
    int i;
    int num_args = ((or_exp*)prog)->num_args;
    if (num_args == 0)
      return make_val_cexp(make_false_val());
    else
    {
      cexp **args = (cexp**)ALLOC(sizeof(cexp*) * num_args);
      for (i = 0; i < num_args; ++i)
	args[i] = compile(((or_exp*)prog)->args[i], e);
      return make_or_cexp(num_args, args);
    }
  }
  case NOT_EXP:
  {
    return make_not_cexp(compile(((not_exp*)prog)->arg, e));
  }
  }
  return NULL;
}

