#define NUM_EXP		0
#define NULL_EXP	1
#define FALSE_EXP	2
#define VOID_EXP	3
#define LAMBDA_EXP	4
#define VAR_EXP		5
#define LET_EXP		6
#define LETREC_EXP	7
#define APP_EXP		8
#define IF_EXP		9
#define BEGIN_EXP	10
#define SETBANG_EXP	11
#define AND_EXP		12
#define OR_EXP		13
#define NOT_EXP		14

typedef struct { int tag; } exp;
typedef struct { int tag; long n; } num_exp;
typedef struct { int tag; } null_exp;
typedef struct { int tag; } false_exp;
typedef struct { int tag; } void_exp;
typedef struct { int tag; int num_parms; char **parms; exp *body; } lambda_exp;
typedef struct { int tag; char *var; } var_exp;
typedef struct { int tag; int num_bindings; char **ids; exp **bound_exps; exp *body; } let_exp;
typedef struct { int tag; int num_bindings; char **ids; exp **bound_exps; exp *body; } letrec_exp;
typedef struct { int tag; exp *fun; int num_args; exp **args; } app_exp;
typedef struct { int tag; exp *test; exp *t_br; exp *f_br; } if_exp;
typedef struct { int tag; int num_exps; exp **exps; } begin_exp;
typedef struct { int tag; char *var; exp *val; } setbang_exp;
typedef struct { int tag; int num_args; exp **args; } and_exp;
typedef struct { int tag; int num_args; exp **args; } or_exp;
typedef struct { int tag; exp *arg; } not_exp;

exp *parse_file(char *f);

static inline exp *make_num_exp(long n)
{
  num_exp *res = (num_exp*)ALLOC(sizeof(num_exp));
  res->tag = NUM_EXP;
  res->n = n;
  return (exp*)res;
}

static inline exp *make_null_exp()
{
  exp *res = (exp*)ALLOC(sizeof(null_exp));
  res->tag = NULL_EXP;
  return (exp*)res;
}

static inline exp *make_false_exp()
{
  exp *res = (exp*)ALLOC(sizeof(false_exp));
  res->tag = FALSE_EXP;
  return (exp*)res;
}

static inline exp *make_void_exp()
{
  exp *res = (exp*)ALLOC(sizeof(void_exp));
  res->tag = VOID_EXP;
  return (exp*)res;
}

static inline exp *make_lambda_exp(int num_parms, char **ids, exp *body)
{
  lambda_exp *res = (lambda_exp*)ALLOC(sizeof(lambda_exp));
  res->tag = LAMBDA_EXP;
  res->num_parms = num_parms;
  res->parms = ids;
  res->body = body;
  return (exp*)res;
}

static inline exp *make_var_exp(char *var)
{
  var_exp *res = (var_exp*)ALLOC(sizeof(var_exp));
  res->tag = VAR_EXP;
  res->var = var;
  return (exp*)res;
}

static inline exp *make_let_exp(int num, char **names, exp **b, exp *body)
{
  let_exp *res = (let_exp*)ALLOC(sizeof(let_exp));
  res->tag = LET_EXP;
  res->num_bindings = num;
  res->ids = names;
  res->bound_exps = b;
  res->body = body;
  return (exp*)res;
}

static inline exp *make_letrec_exp(int num, char **names, exp **b, exp *body)
{
  letrec_exp *res = (letrec_exp*)ALLOC(sizeof(letrec_exp));
  res->tag = LETREC_EXP;
  res->num_bindings = num;
  res->ids = names;
  res->bound_exps = b;
  res->body = body;
  return (exp*)res;
}

static inline exp *make_app_exp(exp *f, int n, exp **a)
{
  app_exp *res = (app_exp*)ALLOC(sizeof(app_exp));
  res->tag = APP_EXP;
  res->fun = f;
  res->num_args = n; 
  res->args = a;
  return (exp*)res;
}

static inline exp *make_if_exp(exp *a, exp *b, exp *c)
{
  if_exp *res = (if_exp*)ALLOC(sizeof(if_exp));
  res->tag = IF_EXP;
  res->test = a;
  res->t_br = b;
  res->f_br = c;
  return (exp*)res;
}

static inline exp *make_begin_exp(int a, exp **b)
{
  begin_exp *res = (begin_exp*)ALLOC(sizeof(begin_exp));
  res->tag = BEGIN_EXP;
  res->num_exps = a;
  res->exps = b;
  return (exp*)res;
}

static inline exp *make_setbang_exp(char *a, exp *e)
{
  setbang_exp *res = (setbang_exp*)ALLOC(sizeof(setbang_exp));
  res->tag = SETBANG_EXP;
  res->var = a;
  res->val = e;
  return (exp*)res;
}

static inline exp *make_and_exp(int a, exp **b)
{
  and_exp *res = (and_exp*)ALLOC(sizeof(and_exp));
  res->tag = AND_EXP;
  res->num_args = a;
  res->args = b;
  return (exp*)res;
}

static inline exp *make_or_exp(int a, exp **b)
{
  or_exp *res = (or_exp*)ALLOC(sizeof(or_exp));
  res->tag = OR_EXP;
  res->num_args = a;
  res->args = b;
  return (exp*)res;
}

static inline exp *make_not_exp(exp *e)
{
  not_exp *res = (not_exp*)ALLOC(sizeof(not_exp));
  res->tag = NOT_EXP;
  res->arg = e;
  return (exp*)res;
}

static inline void print_exp(exp *e);

static inline void print_vars(int i, char **c)
{
  int j = 0;
  for (j = 0; j < i; ++j)
    printf("%s ", c[j]);
}


static inline void print_let(int i, char **ids, exp **bounds)
{
  int j;
  for (j = 0; j < i; ++j)
  {
    printf("(%s ", ids[j]);
    print_exp(bounds[j]);
    printf(")");
  }
}

static inline void print_exps(int i, exp **exps)
{
  int j;
  for (j = 0; j < i; ++j)
  {
    print_exp(exps[j]);
    printf(" ");
  }
}


static inline void print_exp(exp *e)
{
  switch (e->tag)
  {
  case NUM_EXP:
    printf("%ld", ((num_exp*)e)->n);
    break;
  case NULL_EXP:
    printf("null");
    break;
  case FALSE_EXP:
    printf("#f");
    break;
  case VOID_EXP:
    printf("void");
    break;
  case LAMBDA_EXP:
    printf("(lambda (");
    print_vars(((lambda_exp*)e)->num_parms,
	       ((lambda_exp*)e)->parms);
    printf(") ");
    print_exp(((lambda_exp*)e)->body);
    printf(")");
    break;
  case VAR_EXP:
    printf("%s", ((var_exp*)e)->var);
    break;
  case LET_EXP:
    printf("(let (");
    print_let(((let_exp*)e)->num_bindings,
	      ((let_exp*)e)->ids,
	      ((let_exp*)e)->bound_exps);
    printf(")");
    print_exp(((let_exp*)e)->body);
    break;
  case LETREC_EXP:
    printf("(letrec (");
    print_let(((let_exp*)e)->num_bindings,
	      ((let_exp*)e)->ids,
	      ((let_exp*)e)->bound_exps);
    printf(")");
    print_exp(((let_exp*)e)->body);
    break;
  case APP_EXP:
    printf("(");
    print_exp(((app_exp*)e)->fun);
    printf(" ");
    print_exps(((app_exp*)e)->num_args,
	       ((app_exp*)e)->args);
    printf(")");
    break;
  case IF_EXP:
    printf("(if ");
    print_exp(((if_exp*)e)->test);
    printf(" ");
    print_exp(((if_exp*)e)->t_br);
    printf(" ");
    print_exp(((if_exp*)e)->f_br);
    printf(")");
    break;
  case BEGIN_EXP:
    printf("(begin ");
    print_exps(((begin_exp*)e)->num_exps,
	       ((begin_exp*)e)->exps);
    printf(")");
    break;
  case SETBANG_EXP:
    printf("(set! %s ", ((setbang_exp*)e)->var);
    print_exp(((setbang_exp*)e)->val);
    printf(")");
    break;
  case AND_EXP:
    printf("(and ");
    print_exps(((and_exp*)e)->num_args,
	       ((and_exp*)e)->args);
    printf(")");
    break;
  case OR_EXP:
    printf("(or ");
    print_exps(((or_exp*)e)->num_args,
	       ((or_exp*)e)->args);
    printf(")");
    break;
  case NOT_EXP:
    printf("(not ");
    print_exp(((not_exp*)e)->arg);
    printf(")");
    break;
  }
}
