#include "scheme_interp.h"
#include <stdlib.h>
#include <stdio.h>


static inline void binop_error(char *op, val *arg0, val *arg1)
{
  printf("Numeric op(*) given non-numeric arg: ");
  print_value(arg0);
  printf(", ");
  print_value(arg1);
  exit(1);
}
static char *prim_names[] = { "cons", "car", "cdr", "set-car!", "set-cdr!",
                              "null?", "pair?", 
                              "=", "<", ">", "+", "-", "*" };
/* GC Roots */
cexp *exp_arg;
env *env_arg;
cont *cont_arg;
val *val_arg;
val **prim_args;

val *false_const;
val *zero_const;
val *void_const;
val *null_const;

int main(int argc, char **argv)
{  
  heap_val **roots[9] = { (heap_val**)&exp_arg,
                          (heap_val**)&env_arg,
                          (heap_val**)&cont_arg,
                          (heap_val**)&val_arg,
                          (heap_val**)&prim_args,
                          (heap_val**)&false_const,
                          (heap_val**)&zero_const,
                          (heap_val**)&void_const,
                          (heap_val**)&null_const };
  
  memory_init(roots, 9);

  /* Initialize the expression */
  exp_arg = compile(parse_file(argv[1]),
                    make_extended_env(13, prim_names, NULL, make_empty_env()));
  printf("\n");
  /* Initialize the environment */
  {
    val **funcs;
    funcs = (val**)ALLOC(sizeof(val*) * 13);

    funcs[0] = make_prim_func(2, &&cons_prim);
    funcs[1] = make_prim_func(1, &&car_prim);
    funcs[2] = make_prim_func(1, &&cdr_prim);
    funcs[3] = make_prim_func(2, &&setcarbang_prim);
    funcs[4] = make_prim_func(2, &&setcdrbang_prim);
    funcs[5] = make_prim_func(1, &&nullp_prim);
    funcs[6] = make_prim_func(1, &&pairp_prim);
    funcs[7] = make_prim_func(2, &&numeqp_prim);
    funcs[8] = make_prim_func(2, &&numltp_prim);
    funcs[9] = make_prim_func(2, &&numgtp_prim);
    funcs[10] = make_prim_func(2, &&numplus_prim);
    funcs[11] = make_prim_func(2, &&numminus_prim);
    funcs[12] = make_prim_func(2, &&numtimes_prim);
    
    env_arg = make_extended_env(13, NULL, funcs, make_empty_env());
  }

  /* Initialize the continuation */
  cont_arg = make_done_cont();

  prim_args = NULL;
  val_arg = NULL;

 cont_eval_exp:
  switch (exp_arg->tag)
  {
  case VAL_CEXP:
    val_arg = ((val_cexp*)exp_arg)->v;
    goto apply_cont;
  case LAMBDA_CEXP:
    val_arg = make_closure(((lambda_cexp*)exp_arg)->num_parms,
                           ((lambda_cexp*)exp_arg)->body,
                           env_arg);
    goto apply_cont;
  case LEXVAR_CEXP:
  {
    int dist = ((lexvar_cexp*)exp_arg)->dist;
    int offset = ((lexvar_cexp*)exp_arg)->offset;
    env *e = env_arg;
    while (dist > 0)
    {
      e = ((extended_env*)e)->next;
      dist--;
    }
    val_arg = ((extended_env*)e)->vals[offset];
    goto apply_cont;
  }
  case APP_CEXP:
    cont_arg = make_narg_cont(((app_cexp*)exp_arg)->num_args,
                              ((app_cexp*)exp_arg)->args,
                              env_arg,
                              cont_arg);
    exp_arg = ((app_cexp*)exp_arg)->fun;
    goto cont_eval_exp;
  case IF_CEXP:
    cont_arg = make_if_test_cont(((if_cexp*)exp_arg)->t_br,
                                 ((if_cexp*)exp_arg)->f_br,
                                 env_arg,
                                 cont_arg);
    exp_arg = ((if_cexp*)exp_arg)->test;
    goto cont_eval_exp;
  case LEXSETBANG_CEXP:
    cont_arg = make_lexsetbang_cont(((lexsetbang_cexp*)exp_arg)->dist,
                                    ((lexsetbang_cexp*)exp_arg)->offset,
                                    env_arg,
                                    cont_arg);
    exp_arg = ((lexsetbang_cexp*)exp_arg)->val;
    goto cont_eval_exp;
  case AND_CEXP:
    cont_arg = make_and_cont(((and_cexp*)exp_arg)->num_args,
                             1,
                             ((and_cexp*)exp_arg)->args,
                             env_arg,
                             cont_arg);
    exp_arg = ((and_cexp*)exp_arg)->args[0];
    goto cont_eval_exp;
  case OR_CEXP:
    cont_arg = make_or_cont(((or_cexp*)exp_arg)->num_args,
                            1,
                            ((or_cexp*)exp_arg)->args,
                            env_arg,
                            cont_arg);
    exp_arg = (((or_cexp*)exp_arg)->args[0]);
    goto cont_eval_exp;
  case NOT_CEXP:
    cont_arg = make_not_cont(cont_arg);
    exp_arg = ((not_cexp*)exp_arg)->arg;
    goto cont_eval_exp;
  }
 
 apply_cont:
  switch (cont_arg->tag)
  {
  case DONE_CONT:
    print_value(val_arg);
    exit(0);
  case NARG_CONT:
  {
    int num_args = ((narg_cont*)cont_arg)->num_args;
    if (num_args == 0)
    {
      cont_arg = make_farg_cont(val_arg, num_args, 0, NULL,
                                ((narg_cont*)cont_arg)->e,
                                ((narg_cont*)cont_arg)->c);
      val_arg = NULL;
      goto apply_cont;
    }
    else
    {
      int i;
      heap_val **args = (heap_val**)ALLOC(sizeof(heap_val*) * num_args);
      exp_arg = ((narg_cont*)cont_arg)->args[0];
      env_arg = ((narg_cont*)cont_arg)->e;
      for (i = 0; i < num_args; ++i)
        args[i] = (heap_val*)((narg_cont*)cont_arg)->args[i];
      cont_arg = make_farg_cont(val_arg, num_args, 1, args,
                                ((narg_cont*)cont_arg)->e,
                                ((narg_cont*)cont_arg)->c);
      goto cont_eval_exp;
    }
  }
  case FARG_CONT:
  {
    int index = ((farg_cont*)cont_arg)->index;
    int num_args = ((farg_cont*)cont_arg)->num_args;
    heap_val **args = ((farg_cont*)cont_arg)->args;

    if (index > 0)
      args[index - 1] = (heap_val*)val_arg;
    
    if (index == num_args)
    {
      val *fun = ((farg_cont*)cont_arg)->fun;
      switch (fun->tag)
      {
      case CLOSURE:
      {
        if (index == ((closure*)fun)->num_parms)
        {
          exp_arg = ((closure*)fun)->body;
          env_arg = make_extended_env(index, NULL,
                                     (val**)((farg_cont*)cont_arg)->args,
                                     ((closure*)fun)->cenv);
          cont_arg = ((farg_cont*)cont_arg)->c;
          goto cont_eval_exp;
        }
        else
        {
          printf("Function expects %d args, given %d.\n",
                 ((closure*)fun)->num_parms,
                 index);
          exit(1);
        }
      }
      case PRIM_FUNC:
      {
        if (index == ((prim_func*)fun)->num_parms)
        {
          prim_args = (val**)((farg_cont*)cont_arg)->args;
          cont_arg = ((farg_cont*)cont_arg)->c;
          goto *(((prim_func*)fun)->prim_addr);
        }
        else
        {
          printf("Primitive expects %d args, given %d.\n",
                 ((prim_func*)fun)->num_parms,
                 index);
          exit(1);
        }
      }
      default:
        printf("Attempt to apply non-function.\n");
        exit(1);
      }
    }
    else
    {
      exp_arg = (cexp*)args[index];
      env_arg = ((farg_cont*)cont_arg)->e;
      ((farg_cont*)cont_arg)->index = index + 1;
      goto cont_eval_exp;
    }
  }
  case IF_TEST_CONT:
    env_arg = ((if_test_cont*)cont_arg)->e;
    if (val_arg->tag == FALSE_VAL)
      exp_arg = ((if_test_cont*)cont_arg)->f_br;
    else
      exp_arg = ((if_test_cont*)cont_arg)->t_br;
    cont_arg = ((if_test_cont*)cont_arg)->c;
    goto cont_eval_exp;
  case LEXSETBANG_CONT:
  {
    int dist = ((lexsetbang_cont*)cont_arg)->dist;
    int offset = ((lexsetbang_cont*)cont_arg)->offset;
    env *e = ((lexsetbang_cont*)cont_arg)->e;
    while (dist > 0)
    {
      e = ((extended_env*)e)->next;
      dist--;
    }
    ((extended_env*)e)->vals[offset] = val_arg;
    val_arg = void_const;
    cont_arg = ((lexsetbang_cont*)cont_arg)->c;
    goto apply_cont;
  }
  case AND_CONT:
  {
    int index = ((and_cont*)cont_arg)->index;
    if (index == ((and_cont*)cont_arg)->num_args)
    {
      cont_arg = ((and_cont*)cont_arg)->c;
      goto apply_cont;
    }
    else if (val_arg->tag == FALSE_VAL)
    {
      cont_arg = ((and_cont*)cont_arg)->c;
      goto apply_cont;
    }
    else
    {
      env_arg = ((and_cont*)cont_arg)->e;
      ((and_cont*)cont_arg)->index = index + 1;
      exp_arg = ((and_cont*)cont_arg)->args[index];
      goto cont_eval_exp;
    }
  }
  case OR_CONT:
  {
    int index = ((or_cont*)cont_arg)->index;
    if (index == ((or_cont*)cont_arg)->num_args)
    {
      cont_arg = ((or_cont*)cont_arg)->c;
      goto apply_cont;
    }
    else if (val_arg->tag == FALSE_VAL)
    {
      env_arg = ((or_cont*)cont_arg)->e;
      ((or_cont*)cont_arg)->index = index + 1;
      exp_arg = ((or_cont*)cont_arg)->args[index];
      goto cont_eval_exp;
    }
    else
    {
      cont_arg = ((or_cont*)cont_arg)->c;
      goto apply_cont;
    }
  }
  case NOT_CONT:
    if (val_arg->tag == FALSE_VAL)
      val_arg = zero_const;
    else
      val_arg = false_const;
    cont_arg = ((not_cont*)cont_arg)->c;
    goto apply_cont;
  }

 cons_prim:
  val_arg = make_pair_val(prim_args[0], prim_args[1]);
  goto apply_cont;

 car_prim:
  if (prim_args[0]->tag == PAIR_VAL)
  {
    val_arg = ((pair_val*)prim_args[0])->car;
    goto apply_cont;
  }
  else
  {
    printf("car of ");
    print_value(prim_args[0]);
    printf(".\n");
    exit(1);
  }

 cdr_prim:
  if (prim_args[0]->tag == PAIR_VAL)
  {
    val_arg = ((pair_val*)prim_args[0])->cdr;
    goto apply_cont;
  }
  else
  {
    printf("cdr of ");
    print_value(prim_args[0]);
    printf(".\n");
    exit(1);
  }

 setcarbang_prim:
  if (prim_args[0]->tag == PAIR_VAL)
  {
    ((pair_val*)prim_args[0])->car = prim_args[1];
    val_arg = make_void_val();
    goto apply_cont;
  }
  else
  {
    printf("set-car! of ");
    print_value(prim_args[0]);
    printf(".\n");
    exit(1);
  }

 setcdrbang_prim:
  if (prim_args[0]->tag == PAIR_VAL)
  {
    ((pair_val*)prim_args[0])->cdr = prim_args[1];
    val_arg = make_void_val();
    goto apply_cont;
  }
  else
  {
    printf("set-cdr! of ");
    print_value(prim_args[0]);
    printf(".\n");
    exit(1);
  }

 nullp_prim:
  if (prim_args[0]->tag == NULL_VAL)
    val_arg = prim_args[0];
  else
    val_arg = false_const;
  goto apply_cont;

 pairp_prim:
  if (prim_args[0]->tag == NULL_VAL)
    val_arg = prim_args[0];
  else
    val_arg = false_const;
  goto apply_cont;

 numeqp_prim:
  if (prim_args[0]->tag == NUMBER_VAL && prim_args[1]->tag == NUMBER_VAL)
  {
    if (((number_val*)prim_args[0])->n == ((number_val*)prim_args[1])->n)
      val_arg = zero_const;
    else
      val_arg = false_const;
    goto apply_cont;
  }
  else
    binop_error("=", prim_args[0], prim_args[1]);

 numltp_prim:
  if (prim_args[0]->tag == NUMBER_VAL && prim_args[1]->tag == NUMBER_VAL)
  {
    if (((number_val*)prim_args[0])->n < ((number_val*)prim_args[1])->n)
      val_arg = zero_const;
    else
      val_arg = false_const;
    goto apply_cont;
  }
  else
    binop_error("<", prim_args[0], prim_args[1]);

 numgtp_prim:
  if (prim_args[0]->tag == NUMBER_VAL && prim_args[1]->tag == NUMBER_VAL)
  {
    if (((number_val*)prim_args[0])->n > ((number_val*)prim_args[1])->n)
      val_arg = zero_const;
    else
      val_arg = false_const;
    goto apply_cont;
  }
  else
    binop_error(">", prim_args[0], prim_args[1]);

 numplus_prim:
  if (prim_args[0]->tag == NUMBER_VAL && prim_args[1]->tag == NUMBER_VAL)
  {
    val_arg = make_number_val(((number_val*)prim_args[0])->n +
                              ((number_val*)prim_args[1])->n);
    goto apply_cont;
  }
  else
    binop_error("+", prim_args[0], prim_args[1]);

 numminus_prim:
  if (prim_args[0]->tag == NUMBER_VAL && prim_args[1]->tag == NUMBER_VAL)
  {
    val_arg = make_number_val(((number_val*)prim_args[0])->n -
                              ((number_val*)prim_args[1])->n);
    goto apply_cont;
  }
  else
    binop_error("-", prim_args[0], prim_args[1]);

 numtimes_prim:
  if (prim_args[0]->tag == NUMBER_VAL && prim_args[1]->tag == NUMBER_VAL)
  {
    val_arg = make_number_val(((number_val*)prim_args[0])->n *
                              ((number_val*)prim_args[1])->n);
    goto apply_cont;
  }
  else
    binop_error("*", prim_args[0], prim_args[1]);

  return 0;

}
