#ifndef _SCHEME_INTERP_H_
#define _SCHEME_INTERP_H_

#include <stdlib.h>
#include <stdio.h>

extern unsigned char *alloc_ptr;
extern unsigned char *alloc_limit;
extern unsigned char *heap;
void memory_init();
unsigned char *allocate(int bytes);

#define ALLOC(n) allocate(n)

#include "scheme_exp.h"

/* This is used only in the parser.  Since it untagged the GC
   does not understand it.
*/
typedef struct cons_ { void *car; struct cons_ *cdr; } cons_cell;

#define BYTES 1024 * 1024

#define VAL_CEXP	15
#define LAMBDA_CEXP	16
#define LEXVAR_CEXP 	17
#define APP_CEXP	18
#define IF_CEXP		19
#define LEXSETBANG_CEXP	20
#define AND_CEXP	21
#define OR_CEXP		22
#define NOT_CEXP	23

#define EMPTY_ENV	24
#define EXTENDED_ENV	25

#define NUMBER_VAL	26
#define NULL_VAL	27
#define FALSE_VAL	28
#define CLOSURE		29
#define VOID_VAL		30
#define PRIM_FUNC	31
#define PAIR_VAL	32

#define DONE_CONT	33
#define NARG_CONT	34
#define FARG_CONT	35
#define IF_TEST_CONT	36
#define LEXSETBANG_CONT	37
#define AND_CONT	38
#define OR_CONT		39
#define NOT_CONT	40

typedef struct { int tag; } heap_val;

typedef struct { int tag; } cexp;
typedef struct { int tag; } env;
typedef struct { int tag; } val;
typedef struct { int tag; } cont;

typedef struct { int tag; val *v; } val_cexp;
typedef struct { int tag; int num_parms; cexp *body; } lambda_cexp;
typedef struct { int tag; int dist; int offset; } lexvar_cexp;
typedef struct { int tag; int num_args; cexp *fun; cexp **args; } app_cexp;
typedef struct { int tag; cexp *test; cexp *t_br; cexp *f_br; } if_cexp;
typedef struct { int tag; int dist; int offset; cexp *val; } lexsetbang_cexp;
typedef struct { int tag; int num_args; cexp **args; } and_cexp;
typedef struct { int tag; int num_args; cexp **args; } or_cexp;
typedef struct { int tag; cexp *arg; } not_cexp;

typedef struct { int tag; } empty_env;
typedef struct { int tag; int num_bindings; char **syms; val **vals; env *next; } extended_env;

typedef struct { int tag; long n; } number_val;
typedef struct { int tag; } null_val;
typedef struct { int tag; } false_val;
typedef struct { int tag; int num_parms; cexp *body; env *cenv; } closure;
typedef struct { int tag; } void_val;
typedef struct { int tag; int num_parms; void *prim_addr; } prim_func;
typedef struct { int tag; val *car; val *cdr; } pair_val;


typedef struct { int tag; } done_cont;
typedef struct { int tag; int num_args; cexp **args; env *e; cont *c; } narg_cont;
typedef struct { int tag; val *fun; int num_args; int index; heap_val **args; env *e; cont *c; } farg_cont;
typedef struct { int tag; cexp *t_br; cexp *f_br; env *e; cont *c; } if_test_cont;
typedef struct { int tag; int dist; int offset; env *e; cont *c; } lexsetbang_cont;
typedef struct { int tag; int num_args; int index; cexp **args; env *e; cont *c; } and_cont;
typedef struct { int tag; int num_args; int index; cexp **args; env *e; cont *c; } or_cont;
typedef struct { int tag; cont *c; } not_cont;

extern val *false_const;
extern val *zero_const;
extern val *void_const;
extern val *null_const;
cexp *compile(exp *ex, env *en);



static inline cexp *make_val_cexp(val *v)
{
  val_cexp *res = (val_cexp*)ALLOC(sizeof(val_cexp));
  res->tag = VAL_CEXP;
  res->v = v;
  return (cexp*)res;
}

static inline cexp *make_lambda_cexp(int num_parms, cexp *body)
{
  lambda_cexp *res = (lambda_cexp*)ALLOC(sizeof(lambda_cexp));
  res->tag = LAMBDA_CEXP;
  res->num_parms = num_parms;
  res->body = body;
  return (cexp*)res;
}

static inline cexp *make_lexvar_cexp(int d, int o)
{
  lexvar_cexp *res = (lexvar_cexp*)ALLOC(sizeof(lexvar_cexp));
  res->tag = LEXVAR_CEXP;
  res->dist = d;
  res->offset = o;
  return (cexp*)res;
}

static inline cexp *make_app_cexp(cexp *f, int n, cexp **a)
{
  app_cexp *res = (app_cexp*)ALLOC(sizeof(app_cexp));
  res->tag = APP_CEXP;
  res->fun = f;
  res->num_args = n; 
  res->args = a;
  return (cexp*)res;
}

static inline cexp *make_if_cexp(cexp *a, cexp *b, cexp *c)
{
  if_cexp *res = (if_cexp*)ALLOC(sizeof(if_cexp));
  res->tag = IF_CEXP;
  res->test = a;
  res->t_br = b;
  res->f_br = c;
  return (cexp*)res;
}

static inline cexp *make_lexsetbang_cexp(int d, int o, cexp *e)
{
  lexsetbang_cexp *res = (lexsetbang_cexp*)ALLOC(sizeof(lexsetbang_cexp));
  res->tag = LEXSETBANG_CEXP;
  res->dist = d;
  res->offset = o;
  res->val = e;
  return (cexp*)res;
}

static inline cexp *make_and_cexp(int a, cexp **b)
{
  and_cexp *res = (and_cexp*)ALLOC(sizeof(and_cexp));
  res->tag = AND_CEXP;
  res->num_args = a;
  res->args = b;
  return (cexp*)res;
}

static inline cexp *make_or_cexp(int a, cexp **b)
{
  or_cexp *res = (or_cexp*)ALLOC(sizeof(or_cexp));
  res->tag = OR_CEXP;
  res->num_args = a;
  res->args = b;
  return (cexp*)res;
}

static inline cexp *make_not_cexp(cexp *e)
{
  not_cexp *res = (not_cexp*)ALLOC(sizeof(not_cexp));
  res->tag = NOT_CEXP;
  res->arg = e;
  return (cexp*)res;
}

static inline val *make_number_val(long n)
{
  if (n == 0)
    return zero_const;
  else
  {
    number_val *res = (number_val*)ALLOC(sizeof(number_val));
    res->tag = NUMBER_VAL;
    res->n = n;
    return (val*)res;
  }
}

static inline val *make_null_val()
{
  return null_const;
}

static inline val *make_false_val()
{
  return false_const;
}

static inline val *make_void_val()
{
  return void_const;
}

static inline val *make_closure(int n, cexp *b, env *e)
{
  closure *res = (closure*)ALLOC(sizeof(closure));
  res->tag = CLOSURE;
  res->num_parms = n;
  res->body = b;
  res->cenv = e;
  return (val*)res;
}

static inline val *make_prim_func(int n, void *addr)
{
  prim_func *res = (prim_func*)ALLOC(sizeof(prim_func));
  res->tag = PRIM_FUNC;
  res->num_parms = n;
  res->prim_addr = addr;
  return (val*)res;
}

static inline val *make_pair_val(val *car, val *cdr)
{
  pair_val *res = (pair_val*)ALLOC(sizeof(pair_val));
  res->tag = PAIR_VAL;
  res->car = car;
  res->cdr = cdr;
  return (val*)res;
}

static inline env *make_empty_env()
{
  env *e = (env*)ALLOC(sizeof(empty_env));
  e->tag = EMPTY_ENV;
  return e;
}

static inline env *make_extended_env(int i, char **c, val **v, env *ev)
{
  extended_env *e = (extended_env*)ALLOC(sizeof(extended_env));
  e->tag = EXTENDED_ENV;
  e->num_bindings = i;
  e->syms = c;
  e->vals = v;
  e->next = ev;
  return (env*)e;
}

static inline void print_cexp(cexp *e);

static inline void print_cexps(int i, cexp **cexps)
{
  int j;
  for (j = 0; j < i; ++j)
  {
    print_cexp(cexps[j]);
    printf(" ");
  }
}

static inline void print_value(val* v)
{
  switch (v->tag)
  {
  case NUMBER_VAL:
    printf("%ld", ((number_val*)v)->n);
    break;
  case NULL_VAL:
    printf("null");
    break;
  case FALSE_VAL:
    printf("#f");
    break;
  case CLOSURE:
    printf("closure");
    break;
  case VOID_VAL:
    printf("void");
    break;
  case PRIM_FUNC:
    printf("primitive function");
    break;
  case PAIR_VAL:
    printf("(cons ");
    print_value(((pair_val*)v)->car);
    printf(" ");
    print_value(((pair_val*)v)->cdr);
    printf(")");
    break;
  }
}

static inline void print_cexp(cexp *e)
{
  switch (e->tag)
  {
  case VAL_CEXP:
    print_value(((val_cexp*)e)->v);
    break;
  case LAMBDA_CEXP:
    printf("(lambda %d ", ((lambda_cexp*)e)->num_parms);
    print_cexp(((lambda_cexp*)e)->body);
    printf(")");
    break;
  case LEXVAR_CEXP:
    printf("(lexvar %d %d)", ((lexvar_cexp*)e)->dist,
	   ((lexvar_cexp*)e)->offset);
    break;
  case APP_CEXP:
    printf("(");
    print_cexp(((app_cexp*)e)->fun);
    printf(" ");
    print_cexps(((app_cexp*)e)->num_args,
	       ((app_cexp*)e)->args);
    printf(")");
    break;
  case IF_CEXP:
    printf("(if ");
    print_cexp(((if_cexp*)e)->test);
    printf(" ");
    print_cexp(((if_cexp*)e)->t_br);
    printf(" ");
    print_cexp(((if_cexp*)e)->f_br);
    printf(")");
    break;
  case LEXSETBANG_CEXP:
    printf("(lexset! %d %d ", ((lexsetbang_cexp*)e)->dist, 
	   ((lexsetbang_cexp*)e)->offset);
    print_cexp(((lexsetbang_cexp*)e)->val);
    printf(")");
    break;
  case AND_CEXP:
    printf("(and ");
    print_cexps(((and_cexp*)e)->num_args,
	       ((and_cexp*)e)->args);
    printf(")");
    break;
  case OR_CEXP:
    printf("(or ");
    print_cexps(((or_cexp*)e)->num_args,
	       ((or_cexp*)e)->args);
    printf(")");
    break;
  case NOT_CEXP:
    printf("(not ");
    print_cexp(((not_cexp*)e)->arg);
    printf(")");
    break;
  }
}

static inline cont *make_done_cont()
{
  cont *res = (cont*)ALLOC(sizeof(done_cont));
  res->tag = DONE_CONT;
  return res;
}

static inline cont *make_narg_cont(int num, cexp **args, env *e, cont *c)
{
  narg_cont *res = (narg_cont*)ALLOC(sizeof(narg_cont));
  res->tag = NARG_CONT;
  res->num_args = num;
  res->args = args;
  res->e = e;
  res->c = c;
  return (cont*)res;
}

static inline cont *make_farg_cont(val *f, int num, int idx, heap_val **args, env *e, cont *c)
{
  farg_cont *res = (farg_cont*)ALLOC(sizeof(farg_cont));
  res->tag = FARG_CONT;
  res->fun = f;
  res->num_args = num;
  res->index = idx;
  res->args = args;
  res->e = e;
  res->c = c;
  return (cont*)res;
}

static inline cont *make_if_test_cont(cexp *t, cexp *f, env *e, cont *c)
{
  if_test_cont *res = (if_test_cont*)ALLOC(sizeof(if_test_cont));
  res->tag = IF_TEST_CONT;
  res->t_br = t;
  res->f_br = f;
  res->e = e;
  res->c = c;
  return (cont*)res;
}

static inline cont *make_lexsetbang_cont(int d, int o, env *e, cont *c)
{
  lexsetbang_cont *res = (lexsetbang_cont*)ALLOC(sizeof(lexsetbang_cont));
  res->tag = LEXSETBANG_CONT;
  res->dist = d;
  res->offset = o;
  res->e = e;
  res->c = c;
  return (cont*)res;
}

static inline cont *make_and_cont(int n, int i, cexp **a, env *e, cont *c)
{
  and_cont *res = (and_cont*)ALLOC(sizeof(and_cont));
  res->tag = AND_CONT;
  res->num_args = n;
  res->index = i;
  res->args = a;
  res->e = e;
  res->c = c;
  return (cont*)res;
}

static inline cont *make_or_cont(int n, int i, cexp **a, env *e, cont *c)
{
  or_cont *res = (or_cont*)ALLOC(sizeof(or_cont));
  res->tag = OR_CONT;
  res->num_args = n;
  res->index = i;
  res->args = a;
  res->e = e;
  res->c = c;
  return (cont*)res;
}

static inline cont *make_not_cont(cont *c)
{
  not_cont *res = (not_cont*)ALLOC(sizeof(not_cont));
  res->tag = NOT_CONT;
  res->c = c;
  return (cont*)res;
}
#endif
